import logging

import yaml

from .dumper import SpecialDumper as __SpecialDumper
from .loader import (
    TypeOrGeneric as __TypeOrGeneric,
    special_loader as __special_loader
)

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.WARNING)


def dump(data, stream=None, **kwargs):
    """Dump data with yaml.dump with tyaml.SpecialDumper as dumper"""
    _k_dumper = kwargs.pop("Dumper", None)
    if _k_dumper is not None:
        LOGGER.warning("'Dumper' argument will be ignored")  # pragma: nocover
    return yaml.dump(data, stream, Dumper=__SpecialDumper, **kwargs)


def load(stream, as_type: __TypeOrGeneric):
    """Load yaml"""
    return yaml.load(stream, Loader=__special_loader(as_type))



import functools
import inspect
import re
from dataclasses import fields, is_dataclass

MAPPING_RE = re.compile(r"\s+([\w_\d]+?):?\s.*#\s*yaml:\s?(\w+)\n?")


def _get_container_fields(cls):
    # data classes
    if is_dataclass(cls):
        return {field.name: field.name for field in fields(cls)}
    # named tuples
    if hasattr(cls, "_fields"):
        return {field: field for field in getattr(cls, "_fields")}
    return {}


@functools.lru_cache(maxsize=64, typed=True)
def get_mappings(cls) -> dict:
    """Get `class field`: `yaml field` mapping for a class"""
    if inspect.isbuiltin(cls):
        return {}
    _fields = _get_container_fields(cls)

    try:
        src = inspect.getsource(cls)
    except OSError:  # e.g. this happens with collections.namedtuple
        return _fields
    except TypeError:  # in case of built-ins
        return {}

    src_lines = src.split("\n")
    for line in src_lines:
        match = MAPPING_RE.match(line)
        if match is None:
            continue
        _fields[match.group(1)] = match.group(2)
    return _fields



from copy import deepcopy
from dataclasses import is_dataclass
from typing import Any, Type, Union, _GenericAlias, _SpecialForm

from yaml import CollectionNode, FullLoader, MappingNode, SequenceNode

from tyaml.types import get_mappings


def _is_generic(cls):
    if isinstance(cls, _GenericAlias):
        return True

    if isinstance(cls, _SpecialForm):
        return cls not in {Any}

    return False


TypeOrGeneric = Union[type, _GenericAlias]


def _real_cls(cls: TypeOrGeneric) -> type:
    return cls.__origin__


def __kwarg_constructor(loader, node, typ, fld_mappings):
    fields = loader.construct_mapping(node, True)
    kwargs = {cls_field: fields[yml_field] for cls_field, yml_field in fld_mappings.items()}
    return typ(**kwargs)


def _type_tag(typ):
    return f"!!python/object:{typ.__module__}.{typ.__name__}"


def _add_single_cls_loader(typ, loader: Type[FullLoader], base_path: list):
    tag = _type_tag(typ)
    field_mappings = get_mappings(typ)
    if not field_mappings:
        return

    loader.add_constructor(tag, lambda l, n: __kwarg_constructor(l, n, typ, field_mappings.copy()))
    loader.add_path_resolver(tag, base_path[:], MappingNode)

    if hasattr(typ, "__annotations__"):
        # go deeper
        for field, f_typ in typ.__annotations__.items():
            if f_typ in [str, int, float]:
                continue
            el_path = deepcopy(base_path)
            el_path.append(field)
            _add_path_resolvers(f_typ, loader, el_path)


def _add_complex_resolver(typ, loader: Type[FullLoader], base_path: list, node_type: Type[CollectionNode]):
    item_types = [typ]
    base_path = base_path[:]
    base_path.append((node_type, False))
    if hasattr(typ, "__origin__"):
        if typ.__origin__ is Union:  # Union or Optional
            item_types.extend(typ.__args__)
        raise ValueError(f"Only top-level generics are allowed, but {typ} was given")
    if typ in item_types:
        _add_single_cls_loader(typ, loader, base_path)


def _add_path_resolvers(typ: TypeOrGeneric, loader: Type[FullLoader], base_path: list = None):
    if base_path is None:
        base_path = []
    if _is_generic(typ):
        args = typ.__args__
        if not args:
            return  # can't create new path resolver for generic without arguments
        typ = _real_cls(typ)  # convert typing generics to class
        if typ in [list, tuple, set, frozenset]:
            el_type = args[0]  # type: type
            _add_complex_resolver(el_type, loader, base_path, SequenceNode)
        if issubclass(typ, dict):
            el_type = args[1]
            _add_complex_resolver(el_type, loader, base_path, MappingNode)
    if is_dataclass(typ) or hasattr(typ, "_fields"):
        _add_single_cls_loader(typ, loader, base_path)


def special_loader(as_type: type) -> Type[FullLoader]:
    """Construct new loader class supporting current class structure"""

    class TypedLoader(FullLoader):  # pylint: disable=too-many-ancestors
        """Custom loader with typed resolver"""
        ...

    _add_path_resolvers(as_type, TypedLoader)  # we need to add resolver only to the root typed item

    return TypedLoader



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tyaml']

package_data = \
{'': ['*']}

install_requires = \
['pyyaml>=5.3,<6.0']

setup_kwargs = {
    'name': 'pyyaml-typed',
    'version': '0.1.0',
    'description': 'PyYAML dumper/loader using field descriptions in class comments',
    'long_description': '# pyyaml-typed\n[![Build Status](https://travis-ci.org/outcatcher/pyyaml-typed.svg?branch=master)](https://travis-ci.org/outcatcher/pyyaml-typed)\n[![codecov](https://codecov.io/gh/outcatcher/pyyaml-typed/branch/master/graph/badge.svg)](https://codecov.io/gh/outcatcher/pyyaml-typed)\n![GitHub](https://img.shields.io/github/license/outcatcher/pyyaml-typed)\n\nLibrary providing `dump` and `load` functions for pyyaml supporting `go-yaml`-like\ndescription of yaml fields as class comments\n\nDataclasses and named tuples can be used without defining field names.\nField in comment for them will override default class field name\n\nExample:\n\n```python3\nfrom tyaml import dump\n\n@dataclass\nclass Something:\n    my_fld_1: int\n    # or use `yaml:` comment to rename\n    field2: str  # yaml: my_fld_2\n    \noutput = dump([Something(1, "that\'s"), Something(2, "nice")])\n```\n\nwill create yaml:\n```yaml\n- my_fld_1: 1\n  my_fld_2: "that\'s"\n- my_fld_1: 2\n  my_fld_2: "nice"\n```\n\nand in the other direction: \n```python\nfrom typing import List\n\nfrom tyaml import load\n\nloaded = load(yml_str, List[Something])\nloaded == [Something(1, "that\'s"), Something(2, "nice")]\n```\n',
    'author': 'Anton Kachurin',
    'author_email': 'katchuring@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/outcatcher/pyyaml-typed',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
